import logging
from collections import OrderedDict

from .constants import APP_NAME
from .output import CHAPTER

logger = logging.getLogger(APP_NAME)


class ContainerHistory(object):
    def __init__(self):
        self._hist = OrderedDict()

    def add_container_status(self, id, state, message=None):
        debug_text = "container {}\n- id: {}".format(state, id)
        if message:
            debug_text += "\n{}".format(message)
        logger.debug(CHAPTER.format(debug_text))
        self._hist[id] = state

    def remove_container(self, id):
        logger.debug(CHAPTER.format("container removed\n- id: {}".format(id)))
        del self._hist[id]



class ResultImage(object):
    """
    Represent final image built by incubator.
    """

    def __init__(self, image_id, logs=None):
        self._id = image_id
        self._logs = logs or []

    @property
    def id(self):
        return self._id

    @property
    def logs(self):
        return self._logs

    def add_log(self, log):
        self._logs.append(log)

    def __str__(self):
        return "ResultImage<id: {}>".format(str(self._id))


class CommandLog(object):
    """
    Represent logs for one dockerfile command.
    """

    def __init__(self, command, logs=None):
        self._command = command
        self._logs = logs or []

    @property
    def command(self):
        return self._command

    @property
    def logs(self):
        return self._logs

    def add_log(self, log):
        self._logs.append(log)

    def add_logs(self, logs):
        self._logs += logs



import logging
import os
import tarfile

from .constants import APP_NAME
from .fileutils import (copy_files_to_container, create_directory,
                        norm_path_with_work_dir, put_archive)
from .layer import ImageLayer
from .output import CHAPTER, ERROR_SUBTITLE
from .utilclasses import CommandLog
from .utils import substitute_variables

logger = logging.getLogger(APP_NAME)


class Command:
    def __init__(self, value, content, layer_config, build_context):
        self.value = value
        self.layer_config = layer_config
        self._content = content
        self._build_context = build_context
        self._log = CommandLog(command=self._content)
        self.layer_config.add_log(self._log)

    @staticmethod
    def changes_data():
        pass

    def apply(self, client, config):
        logger.info(CHAPTER.format(self._content))
        self.value = substitute_variables(text=self.value,
                                          variables=self.layer_config.envs)

    def __str__(self):
        return "{} {}".format(self.__class__.__name__, self.value)


class Arg(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)
        value_split = self.value.split('=')

        if len(value_split) == 1:
            key, default_value = value_split[0], ""
        elif len(value_split) == 2:
            key, default_value = value_split
        else:
            raise Exception("Wrong format of ARG command.")

        self.layer_config.envs[key] = config.buildargs.get(key, default_value)
        config.mark_used_buildarg(key=key)


class Copy(Command):
    @staticmethod
    def changes_data():
        return True

    def apply(self, client, config):
        Command.apply(self, client, config)

        splited_command = self.value.split()
        if len(splited_command) < 2:
            logger.warning('Wrong command: {}'.format(str(self.value)))
            return
            # exception?

        source_files = splited_command[0:-1:]
        target_path = norm_path_with_work_dir(path=splited_command[-1],
                                              workdir=self.layer_config.workdir)
        copy_files_to_container(container=self.layer_config.current_container,
                                client=client,
                                config=config,
                                context=self._build_context,
                                source=source_files,
                                target=target_path)


class Add(Command):
    @staticmethod
    def changes_data():
        return True

    def apply(self, client, config):
        Command.apply(self, client, config)

        splited_command = self.value.split()
        if len(splited_command) < 2:
            logger.warning('Wrong command: {}'.format(str(self.value)))
            return
            # exception?

        source_files = splited_command[0:-1:]
        target_path = norm_path_with_work_dir(path=splited_command[-1],
                                              workdir=self.layer_config.workdir)

        for src in source_files:
            file = self._build_context.get_file(src)
            if file:
                try:
                    with tarfile.open(fileobj=file):
                        put_archive(client=client,
                                    config=config,
                                    container=self.layer_config.current_container,
                                    tar=file,
                                    target=target_path)
                        break
                except:
                    pass
            copy_files_to_container(container=self.layer_config.current_container,
                                    client=client,
                                    config=config,
                                    context=self._build_context,
                                    source=[src],
                                    target=target_path)


class Cmd(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)
        self.layer_config.add_property("Cmd", self.value)


class Env(Command):
    @staticmethod
    def changes_data():
        return False


class Expose(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)
        self.layer_config.add_to_dictionary("ExposedPorts", self.value, {})


class From(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)


class Label(Command):
    @staticmethod
    def changes_data():
        return False


class Maintainer(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)
        self.layer_config.author = self.value


class Run(Command):
    @staticmethod
    def changes_data():
        return True

    def apply(self, client, config):
        Command.apply(self, client, config)

        if self.layer_config.workdir and self.layer_config.workdir != os.curdir:
            command = "sh -c \"cd {}; {}\"".format(self.layer_config.workdir, self.value)
        else:
            command = self.value

        for l in client.execute(container=self.layer_config.current_container,
                                cmd=command):
            logger.debug(str(l.decode().rstrip()))
            self._log.add_log(l)
        logger.debug("")


class User(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)
        self.layer_config.add_property("User", self.value)


class Workdir(Command):
    @staticmethod
    def changes_data():
        return False

    def apply(self, client, config):
        Command.apply(self, client, config)
        self.layer_config.workdir = self.value
        create_directory(client=client,
                         config=config,
                         container=self.layer_config.current_container,
                         target=self.layer_config.workdir)


COMMANDS = {
    "ADD": Add,
    "ARG": Arg,
    "CMD": Cmd,
    "COPY": Copy,
    "ENV": Env,
    "EXPOSE": Expose,
    "FROM": From,
    # "LABEL": Label,
    "MAINTAINER": Maintainer,
    "RUN": Run,
    "USER": User,
    "WORKDIR": Workdir
}


def get_layers_from_dockerfile_structure(structure, layers_split_mark, build_context):
    if layers_split_mark is None or layers_split_mark == []:
        layers_split_mark = [i for i in range(len(structure))]
    layers = []
    current_command = 0
    layer_number = 1
    layer = ImageLayer(number=0)
    data_changed = False
    for layer_split in layers_split_mark + [len(structure)]:

        for i in range(current_command, layer_split):
            cmd = structure[i]
            instruction = cmd["instruction"]
            cmdType = COMMANDS.get(instruction)
            if not cmdType:
                logger.warning(ERROR_SUBTITLE.format("Unknown instruction {}.".format(instruction)))
            else:
                layer.add_command(command=cmdType(value=cmd["value"],
                                                  content=cmd['content'],
                                                  layer_config=layer.conf,
                                                  build_context=build_context),
                                  content=cmd['content'])

                current_command += 1
                data_changed = data_changed or cmdType.changes_data()

        if data_changed or layer_split == len(structure):
            layers.append(layer)
            if not layer_split == len(structure):
                layer = ImageLayer(number=layer_number)
                layer_number += 1
            else:
                layer.is_last = True

    return layers



"""
The MIT License (MIT)

Copyright (c) 2017 Frantisek Lachman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

Python API for the alternative docker builder.
"""
import logging

from .core import utils
from .core.build import Builder
from .core.caching import ImageCache
from .core.client import DockerClient
from .core.config import ImageConfig
from .core.constants import APP_NAME
from .core.context import BuildContext
from .core.output import ERROR_SUBTITLE

logger = logging.getLogger(APP_NAME)


def build(path=None, fileobj=None, tag=None,
          custom_context=False, pull=False, rm=False,
          forcerm=False, dockerfile=None, buildargs=None,
          container_limits=None, volumes=None,
          labels=None, config=None, context_file_limit=None,
          container_client=None, cache=None):
    """

    :param path: (str) Path to the directory containing the Dockerfile.
    :param fileobj: A file object to use as the Dockerfile. (Or a file-like
                object)
    :param tag: (str or list) A tag to add to the final image.
    :param custom_context: (bool) Optional if using ``fileobj``.
    :param pull: (bool) Downloads any updates to the FROM image in Dockerfiles.
    :param rm: (bool) Remove intermediate containers. The ``docker build``
                command now defaults to ``--rm=true``, but we have kept the old
                default of `False` to preserve backward compatibility.
    :param forcerm: (bool) Always remove intermediate containers,
                        even after unsuccessful builds
    :param dockerfile: (str) A path within the build context to the Dockerfile.
    :param buildargs:  (dict) A dictionary of build arguments.
    :param container_limits: (dict) A dictionary of limits applied to each
                container created by the build process. Valid keys:

                - memory (int): set memory limit for build
                - memswap (int): Total memory (memory + swap), -1 to disable
                    swap
                - cpushares (int): CPU shares (relative weight)
                - cpusetcpus (str): CPUs in which to allow execution, e.g.,
                    ``"0-3"``, ``"0,1"``
    :param labels: (dict) A dictionary of labels to set on the image.
    :param volumes: (dict) A dictionary of bind volumes.
    :param config: (dict or list of dicts) A dictionary of addition configuration.
    :param context_file_limit: (int) Maximum size of in memory config. (0 is unlimited and default)
    :param cache: object responsible for caching images
    :param container_client: client representing basic API
    :return: id of created image
    """

    try:
        container_client = container_client or DockerClient()
    except Exception as ex:
        logger.warning(ERROR_SUBTITLE.format("Cannot launch container client:\n{}".format(ex)))
        raise ex

    cache = cache or ImageCache()

    try:
        config_object = ImageConfig.merge_configs(config)
    except Exception as ex:
        logger.warning(ERROR_SUBTITLE.format("Cannot merge config files:\n{}".format(ex)))
        raise ex

    try:
        config_object.update(buildargs=buildargs,
                             container_limits=container_limits,
                             context_file_limit=context_file_limit,
                             forcerm=forcerm,
                             labels=labels,
                             pull=pull,
                             rm=rm,
                             tags=tag,
                             volumes=volumes)
    except Exception as ex:
        logger.warning(ERROR_SUBTITLE.format("Cannot update config with arguments:\n{}".format(ex)))
        raise ex

    try:
        context_fileobj, dockerfile_fileobj = utils.mkbuildcontext(path=path,
                                                                   fileobj=fileobj,
                                                                   custom_context=custom_context,
                                                                   dockerfile=dockerfile,
                                                                   client=container_client,
                                                                   limit=config_object.context_file_limit)
        build_context = BuildContext(context=context_fileobj, limit=context_file_limit)

    except Exception as ex:
        logger.warning(ERROR_SUBTITLE.format("Cannot make build context:\n{}".format(ex)))
        raise ex

    try:
        builder = Builder(context=build_context,
                          dockerfile=dockerfile_fileobj,
                          client=container_client,
                          config=config_object,
                          cache=cache)
        return builder.build()
    except Exception as ex:
        logger.warning(ERROR_SUBTITLE.format("Build failed:\n{}".format(ex)))
        raise ex



import os

from .constants import COMMIT_MESSAGE_START


class ImageLayer(object):
    """
    Represent one layer in final image.

    Contains commands and configuration for this layer.
    """

    def __init__(self, number, commands=None):
        self._commands = commands or []
        self.number = number
        self.is_last = False
        self.commit_message = COMMIT_MESSAGE_START
        self.conf = LayerConfig()

    def __str__(self):
        commands = "["
        for cmd in self.commands:
            commands += "<{}>".format(str(cmd))
        commands += "]"
        return "layer {}: {}".format(str(self.number), commands)

    def add_command(self, command, content):
        self._commands.append(command)

        formated_content = " {}".format(content.rstrip())
        self.commit_message += formated_content
        label_name = "layer_{}_commands".format(str(self.number))
        self.conf.labels.setdefault(label_name, "")
        self.conf.labels[label_name] += formated_content

    @property
    def commands(self):
        return self._commands


class LayerConfig(object):
    """
    Represent configuration of one layer.

    Contains metadata and commit config for new layer.
    """
    def __init__(self, envs=None, labels=None):
        self.container = None
        self.author = None
        self.labels = labels or {}

        self._logs = []
        self._workdir = os.curdir
        self._envs = envs or {}

        self._commit_config_properties = {}

    def add_property(self, key, value):
        self._commit_config_properties[key] = value

    def add_labels(self, labels):
        self.labels.update(labels)

    def add_envs(self, envs):
        self._envs.update(envs)

    def add_log(self, log):
        self._logs.append(log)

    @property
    def workdir(self):
        return self._workdir

    @workdir.setter
    def workdir(self, workdir):
        new_workdir = os.path.normpath(os.path.join(self._workdir, workdir))
        self.add_property("WorkingDir", new_workdir)
        self._workdir = new_workdir

    @property
    def commit_config(self):
        """
        Read-only property which can be used as a commit-config dictionary
        on container commit.

        :return: dict as a commit-config for this layer
        """
        config = dict(self._commit_config_properties)

        config.setdefault("Env", [])
        for k, v in self._envs.items():
            config["Env"].append('{}:{}'.format(str(k), str(v)))

        config.setdefault("Labels", {})
        config["Labels"].update(self.labels)

        return config

    @property
    def logs(self):
        return self._logs

    @property
    def envs(self):
        return self._envs



import ast
import logging
import re
from distutils.version import StrictVersion

import docker
from docker.errors import APIError
from docker.utils import exclude_paths

from .constants import APP_NAME, DEFAULT_INFINITE_COMMAND, MAX_API_ATTEMPTS
from .history import ContainerHistory
from .output import CHAPTER
from .utils import get_name_and_tag, volumes_to_bind_dict, volumes_to_list

IS_OLD_API = StrictVersion(docker.__version__) < StrictVersion('2.0.0')

logger = logging.getLogger(APP_NAME)


class Client:
    """
    Abstract container client representing basic functionality of container engine.
    """

    def __init__(self, history=None):
        self.history = history or ContainerHistory()

    def build_image(self, fileobj=None, dockerfile=None, custom_context=None, buildargs=None, path=None):
        """
        Build image from given dockerfile, context and build arguments.
        :param fileobj: file-like object representing dockerfile or archive with context
        :param dockerfile: path to dockerfile in build context
        :param custom_context: whether to use fileobj as context
        :param buildargs: build arguments passed to build
        :param path: directory with build context
        :return: id of created image
        """
        pass

    def run_container_infinitely(self, image, infinite_command=None, volumes=None, container_limits=None):
        """
        Run a container with infinite command.
        Container is created, never stop and can be changed via exec.
        :param image: image used for new container
        :param infinite_command: infinite_command can be overwritten (default is in constants.py)
        :param volumes: volumes mounted to container
        :param container_limits: (dict) A dictionary of limits applied to each container created by the build process. Valid keys:
                    memory (int): set memory limit for build
                    memswap (int): Total memory (memory + swap), -1 to disable swap
                    cpushares (int): CPU shares (relative weight)
                    cpusetcpus (str): CPUs in which to allow execution, e.g., "0-3", "0,1"
        :return: id of created container
        """
        pass

    def run_container(self, image, command):
        """
        Run a container with givent command.
        :param image: container image to run
        :param command: command to run in container
        :return: output from command
        """
        pass

    def commit_container(self, container, conf, author, message, tags=None):
        """
        Create new image from container.
        :param container: id of container to commit
        :param conf: configuration forced to client
        :param author: author of commit
        :param message: commit message
        :param tags: (list of tuples) tags connected to new image
        :return: id of created image
        """
        pass

    def execute(self, container, cmd):
        """
        Execute command on running container.
        :param container: id of container to execute command in
        :param cmd: command to execute
        :return: output of executed command as a stream
        """
        pass

    def put_archive(self, container, path, data):
        """
        Put archive into container.
        :param container: id of container to put data in
        :param path: where to put data
        :param data: archive with data
        :return: None
        """
        pass

    def remove_container(self, container):
        """
        Remove container.
        :param container: id of container to remove
        :return: None
        """
        pass

    def remove_image(self, image):
        """
        Remove image.
        :param image: Id of image to remove.
        :return: None
        """
        pass

    def get_image(self, name):
        """
        Get id of image with given name.
        :param name: name or id of image
        :return: id of image with given name or None
        """
        pass

    def pull_image(self, name):
        """
        Pull image with given name from repository.
        :param name: Name of image to pull. Default tag is 'latest'.
        :return: Id of pulled image or None.
        """
        pass

    @staticmethod
    def exclude_paths_from_context(root, exclude, dockerfile):
        """
        :param root: path of build context
        :param exclude: list of .dockerignore patterns
        :param dockerfile: path to dockerfile
        :return: an iterator of all paths (files and directories) not excluded with patterns
                    All paths returned are relative to the root.
        """
        pass


class DockerClient(Client):
    """
    Container client representing docker-py independently to version.
    """

    def __init__(self, history=None):
        Client.__init__(self, history)
        self.client = docker.from_env(version='auto')

    def build_image(self, fileobj=None, dockerfile=None, custom_context=None, buildargs=None, path=None):
        if IS_OLD_API:
            resp = [line for line in self.client.build(rm=True, fileobj=fileobj,
                                                       dockerfile=dockerfile,
                                                       custom_context=custom_context,
                                                       buildargs=buildargs,
                                                       path=path
                                                       )]
            event = ast.literal_eval(resp[-1].decode())
            if 'stream' in event:
                match = re.search(r'Successfully built ([0-9a-f]+)',
                                  event.get('stream', ''))
                if match:
                    image_id = match.group(1)
                    return image_id
        else:
            image = self.client.images.build(rm=True, fileobj=fileobj,
                                             dockerfile=dockerfile,
                                             custom_context=custom_context,
                                             buildargs=buildargs,
                                             path=path)
            return image.id

    def run_container_infinitely(self, image, infinite_command=None, volumes=None, container_limits=None):
        infinite_command = infinite_command or DEFAULT_INFINITE_COMMAND
        volumes = volumes or {}
        container_limits = container_limits or {}

        if IS_OLD_API:
            client = self.client
        else:
            client = self.client.api

        volumes_list = volumes_to_list(volumes, mode="z")
        container_id = client.create_container(image=image,
                                               command=infinite_command,
                                               host_config={"binds": volumes_list,
                                                            "Memory": container_limits.get("memory"),
                                                            'MemorySwap': container_limits.get("memswap"),
                                                            'CpuShares': container_limits.get("cpushares"),
                                                            'CpusetCpus': container_limits.get("cpusetcpus")
                                                            },
                                               stdin_open=True).get('Id')
        self.history.add_container_status(id=container_id,
                                          state="created")
        client.start(container_id)
        self.history.add_container_status(id=container_id,
                                          state="running")
        return container_id

    def run_container(self, image, command, remove=False, volumes=None, container_limits=None):
        volumes = volumes or {}
        container_limits = container_limits or {}
        if IS_OLD_API:
            volumes_list = volumes_to_list(volumes, mode="z")
            container_id = self.client.create_container(image=image,
                                                        command=command,
                                                        host_config={"binds": volumes_list,
                                                                     "Memory": container_limits.get("memory"),
                                                                     'MemorySwap': container_limits.get("memswap"),
                                                                     'CpuShares': container_limits.get("cpushares"),
                                                                     'CpusetCpus': container_limits.get("cpusetcpus")
                                                                     },
                                                        ).get('Id')
            self.history.add_container_status(id=container_id,
                                              state="created")
            self.client.start(container_id)
            self.history.add_container_status(id=container_id,
                                              state="running")

            # TODO: temporary workaround !!!
            self.client.wait(container_id)

            for l in self.client.logs(container=container_id,
                                      stream=True, follow=True,
                                      stdout=True, stderr=True,
                                      tail='all'):
                yield l

            self.history.add_container_status(id=container_id,
                                              state="stopped")
            if remove:
                self.client.remove_container(container_id, force=True)
                self.history.remove_container(id=container_id)

        else:
            volumes_list = volumes_to_bind_dict(volumes, mode="z")
            container = self.client.containers.create(image=image,
                                                      command=command,
                                                      volumes=volumes_list,
                                                      mem_limit=container_limits.get("memory"),
                                                      memswap_limit=container_limits.get("memswap"),
                                                      cpu_shares=container_limits.get("cpushares"),
                                                      cpuset_cpus=container_limits.get("cpusetcpus"))
            self.history.add_container_status(id=container.id,
                                              state=container.status)

            container.start()

            self.history.add_container_status(id=container.id,
                                              state=container.status)

            # TODO: temporary workaround !!!
            container.wait()

            for l in container.logs(stream=True, follow=True, stdout=True, stderr=True, tail='all'):
                yield l

            self.history.add_container_status(id=container.id,
                                              state="stopped")

            if remove:
                container.remove(force=True)
                self.history.remove_container(container.id)

    def commit_container(self, container, conf, author, message, tags=None):
        tags = tags or []
        if IS_OLD_API:
            image_id = self.client.commit(container=container,
                                          conf=conf,
                                          author=author,
                                          message=message).get('Id')
            self.history.add_container_status(container, "committed", "- image: {}".format(image_id))

            for repo, tag in tags:
                self.client.tag(image=image_id,
                                repository=repo,
                                tag=tag)
                logger.debug(
                    CHAPTER.format("image tagged:\n- id: {}\n- repo: {}\n- tag: {}".format(image_id, repo, tag)))
            return image_id
        else:
            container_obj = self.client.containers.get(container)
            image = container_obj.commit(conf=conf,
                                         author=author,
                                         message=message)
            self.history.add_container_status(container, "committed", "- image: {}".format(image.id))
            for repo, tag in tags:
                image.tag(repo, tag)
                logger.debug(
                    CHAPTER.format("image tagged:\n- id: {}\n- repo: {}\n- tag: {}".format(image.id, repo, tag)))
            return image.id

    def execute(self, container, cmd):
        if IS_OLD_API:
            command_id = self.client.exec_create(container=container,
                                                 cmd=cmd).get('Id')
            output = self.client.exec_start(exec_id=command_id,
                                            stream=True)
        else:
            container_obj = self.client.containers.get(container)
            output = container_obj.exec_run(cmd=cmd,
                                            stream=True)
        return output

    def put_archive(self, container, path, data):
        for i in range(MAX_API_ATTEMPTS):
            try:
                if IS_OLD_API:
                    self.client.put_archive(container=container,
                                            path=path,
                                            data=data)
                else:
                    container_obj = self.client.containers.get(container)
                    container_obj.put_archive(path=path, data=data)
            except APIError as exc:
                if i == MAX_API_ATTEMPTS:
                    raise exc

    def stop_container(self, container):
        if IS_OLD_API:
            self.client.stop(container)
            self.history.add_container_status(container, "stopped")
        else:
            container_obj = self.client.containers.get(container)
            container_obj.kill()
            self.history.add_container_status(container, "stopped")

    def remove_container(self, container):
        if IS_OLD_API:
            self.client.remove_container(container)
            self.history.remove_container(container)
        else:
            container_obj = self.client.containers.get(container)
            container_obj.remove()
            self.history.remove_container(container)

    def remove_image(self, image):
        if IS_OLD_API:
            self.client.remove_image(image)
        else:
            image_obj = self.client.images.get(image)
            image_obj.remove()

    def get_image(self, name):
        try:
            if IS_OLD_API:
                image = self.client.inspect_image(name)
                if image:
                    return image.get('Id')
            else:
                image = self.client.images.get(name)
                if image:
                    return image.id
        except APIError:
            return None
        return None

    def pull_image(self, name):
        image_name, image_tag = get_name_and_tag(name)

        if IS_OLD_API:
            self.client.pull(repository=image_name,
                             tag=image_tag)
            return self.get_image(name)
        else:
            image = self.client.images.pull("{}:{}".format(image_name, image_tag))
            if image:
                return image.id
        return None

    @staticmethod
    def exclude_paths_from_context(root, exclude, dockerfile):
        return exclude_paths(root, exclude, dockerfile)



import click


class KeyValue(click.ParamType):
    name = 'key=value'

    def convert(self, value, param, ctx):
        value_split = value.split('=')
        if len(value_split) != 2:
            self.fail("{} is not a valid key=value".format(value), param, ctx)
        key, value = value_split
        return key, value


KEY_VALUE = KeyValue()



"""
Constants used by builder.
"""
import os

CONFIG_BUILDARGS = 'buildargs'
CONFIG_CONTAINER_LIMITS = 'container_limits'
CONFIG_CONTEXT_FILE_LIMIT = "context_file_limit"
CONFIG_FORCE_RM = "forcerm"
CONFIG_INFINITE_COMMAND = "infinite_command"
CONFIG_LABELS = "labels"
CONFIG_LAYERS = "layers"
CONFIG_MKDIR_COMMAND = 'mkdir_command'
CONFIG_NAME = "config_name"
CONFIG_PULL = "pull"
CONFIG_RM = "rm"
CONFIG_TAGS = "tags"
CONFIG_VERSION = "version"
CONFIG_VOLUMES = "volumes"

DEFAULT_CONTEXT_FILE_LIMIT = 0
DEFAULT_REPOSITORY = "incubator-image"
DEFAULT_FORCE_RM = False
DEFAULT_INFINITE_COMMAND = "cat -"
DEFAULT_MKDIR_COMMAND = 'mkdir --parent'
DEFAULT_PULL = False
DEFAULT_RM = False
DEFAULT_TAG = "latest"
DEFAULT_VERSION = 1

COMMIT_MESSAGE_START = "This layer is composed of these commands:"

FILE_MODE_FROM_URL = 0o600

APP_NAME = 'incubator'

CONTEXT_FILE_IN_MEMORY = "in memory"
CONTEXT_FILE_ON_DISK = "on disk"

MAX_API_ATTEMPTS = 10

DEFAULT_CONFIG_FILES = ["{}/.incubator.rc".format(os.path.expanduser("~")),
                        "./.incubator.rc"]



#!/usr/bin/python
"""
The MIT License (MIT)

Copyright (c) 2017 Frantisek Lachman

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from __future__ import print_function

import os
import pkgutil
import re
from distutils.version import StrictVersion

from setuptools import find_packages, setup

description = """Alternative image builder for docker containers.
Provides python2/python3 api and cli tool `incubator`.

Extends the functionality of basic `docker build` with precise layering,
ability to have build time volumes and some other useful options,
that can be given as a python dictionary or json file.
"""


def _get_version():
    _dirname = os.path.dirname(os.path.abspath(__file__))
    _version_file = open(os.path.join(_dirname, 'VERSION'))
    return _version_file.read().strip()


def _get_requirements(path):
    try:
        with open(path) as f:
            packages = f.read().splitlines()
    except (IOError, OSError) as ex:
        raise RuntimeError("Can't open file with requirements: %s", repr(ex))
    return [p.strip() for p in packages if not re.match(r"^\s*#", p)]


def _requires():
    return [r.replace('-', '_').split(">")[0] for r in _install_requirements()]


def _test_module(mod):
    return pkgutil.find_loader(mod) is not None


def _install_requirements():
    requirements = _get_requirements('requirements.txt')

    if _test_module("docker"):
        import docker
        is_old_api = StrictVersion(docker.__version__) < StrictVersion('2.0.0')
        if is_old_api:
            requirements.remove("docker")
            requirements.append("docker-py")
    return requirements


setup(
    name='incubator',
    version=_get_version(),
    description='Python library for building container images.',
    long_description=description,
    author='Frantisek Lachman',
    author_email='lachmanfrantisek@gmail.com',
    url='https://gitlab.com/lachmanfrantisek/incubator',
    license="MIT",
    packages=find_packages(exclude=["tests"]),
    install_requires=_install_requirements(),
    requires=_requires(),
    tests_require=_get_requirements('tests/requirements.txt'),
    entry_points='''
       [console_scripts]
       incubator=incubator.cli.incubator:cli
   ''',
    keywords='containers, build, docker, images',
    classifiers=['Development Status :: 4 - Beta',
                 'Environment :: Console',
                 'Intended Audience :: Developers',
                 'Intended Audience :: System Administrators',
                 'License :: OSI Approved :: MIT License',
                 'Operating System :: POSIX :: Linux',
                 'Programming Language :: Python',
                 'Programming Language :: Python :: 2.7',
                 'Programming Language :: Python :: 3.4',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: Implementation :: CPython',
                 'Topic :: Software Development :: Build Tools',
                 'Topic :: Software Development :: Libraries :: Python Modules',
                 'Topic :: System :: Software Distribution',
                 ]
)



"""
Class for managing configuration.
"""
import json
import logging
import os

import jsonschema
import six
from jsonschema import Draft3Validator

from .constants import (APP_NAME, CONFIG_BUILDARGS, CONFIG_CONTAINER_LIMITS,
                        CONFIG_CONTEXT_FILE_LIMIT, CONFIG_FORCE_RM,
                        CONFIG_INFINITE_COMMAND, CONFIG_LABELS, CONFIG_LAYERS,
                        CONFIG_MKDIR_COMMAND, CONFIG_NAME, CONFIG_PULL,
                        CONFIG_RM, CONFIG_TAGS, CONFIG_VERSION, CONFIG_VOLUMES,
                        DEFAULT_CONFIG_FILES, DEFAULT_CONTEXT_FILE_LIMIT,
                        DEFAULT_FORCE_RM, DEFAULT_INFINITE_COMMAND,
                        DEFAULT_MKDIR_COMMAND, DEFAULT_PULL, DEFAULT_RM,
                        DEFAULT_TAG, DEFAULT_VERSION)

logger = logging.getLogger(APP_NAME)

schema_v1 = {
    "type": "object",
    "properties": {
        CONFIG_VERSION: {"type": "number"},
        CONFIG_NAME: {"type": "string"},
        CONFIG_BUILDARGS: {
            "type": "object",
            "patternProperties": {
                "": {
                    "type": "string"
                }
            }
        },
        CONFIG_CONTAINER_LIMITS: {
            "type": "object",
            "properties": {},
            "patternProperties": {
                "": {
                    "type": "string"
                }
            }
        },
        CONFIG_CONTEXT_FILE_LIMIT: {'type': 'integer'},
        CONFIG_FORCE_RM: {'type': 'boolean'},
        CONFIG_LAYERS: {
            "type": "array",
            "items": {
                "type": "integer",
                "minimum": 0
            },
        },
        CONFIG_VOLUMES: {
            "type": "array",
            "properties": {},
            "patternProperties": {
                "": {
                    "type": "string"
                }
            }
        },
        CONFIG_LABELS: {
            "type": "object",
            "properties": {},
            "patternProperties": {
                "": {
                    "type": "string"
                }
            }
        },
        CONFIG_INFINITE_COMMAND: {"type": "string"},
        CONFIG_MKDIR_COMMAND: {"type": "string"},
        CONFIG_PULL: {"type": "boolean"},
        CONFIG_RM: {"type": "boolean"},
        CONFIG_TAGS: {
            "type": "array",
            "items": {
                "type": "string"
            }
        }
    },
    "additionalProperties": False
}

schema = {1.0: schema_v1}


class ImageConfig(object):
    """
    Representation of config file.
    Structure of file is described and validated by jsonschema.
    """

    def __init__(self, config_dict=None):

        config_dict = config_dict or {}

        if not self.validate_json_schema(config=config_dict):
            raise Exception("Config not valid.")

        name_str = config_dict.get(CONFIG_NAME)
        self._name = [name_str] if name_str else []
        self._buildargs = config_dict.get(CONFIG_BUILDARGS)
        self._container_limits = config_dict.get(CONFIG_CONTAINER_LIMITS)
        self._context_file_limit = config_dict.get(CONFIG_CONTEXT_FILE_LIMIT)
        self._forcerm = config_dict.get(CONFIG_FORCE_RM)
        self._infinite_command = config_dict.get(CONFIG_INFINITE_COMMAND)
        self._labels = config_dict.get(CONFIG_LABELS)

        layers = config_dict.get(CONFIG_LAYERS)
        if layers is not None:
            self._layers = sorted(set(layers))
        else:
            self._layers = None

        self._mkdir_command = config_dict.get(CONFIG_MKDIR_COMMAND)
        self._pull = config_dict.get(CONFIG_PULL)
        self._rm = config_dict.get(CONFIG_RM)
        self._tags = config_dict.get(CONFIG_TAGS)
        self._volumes = config_dict.get(CONFIG_VOLUMES)
        self._unused_buildargs = None

    @property
    def config(self):
        d = {}
        if self._buildargs is not None:
            d[CONFIG_BUILDARGS] = dict(self._buildargs)
        if self._container_limits is not None:
            d[CONFIG_CONTAINER_LIMITS] = dict(self._container_limits)
        if self._context_file_limit:
            d[CONFIG_CONTEXT_FILE_LIMIT] = self._context_file_limit
        if self._forcerm is not None:
            d[CONFIG_FORCE_RM] = self._forcerm
        if self._infinite_command:
            d[CONFIG_INFINITE_COMMAND] = self._infinite_command
        if self._labels is not None:
            d[CONFIG_LABELS] = dict(self._labels)
        if self._layers is not None:
            d[CONFIG_LAYERS] = list(self._layers)
        if self._mkdir_command:
            d[CONFIG_MKDIR_COMMAND] = self._mkdir_command
        if self._pull is not None:
            d[CONFIG_PULL] = self._pull
        if self._rm is not None:
            d[CONFIG_RM] = self._rm
        if self._tags is not None:
            d[CONFIG_TAGS] = list(self._tags)
        if self._volumes is not None:
            d[CONFIG_VOLUMES] = list(self._volumes)

        return d

    @property
    def buildargs(self):
        return self._buildargs or {}

    @buildargs.setter
    def buildargs(self, value):
        self.build_args = value

    @property
    def container_limits(self):
        return self._container_limits or {}

    @container_limits.setter
    def container_limits(self, limits):
        self._container_limits = limits

    @property
    def context_file_limit(self):
        return self._context_file_limit or DEFAULT_CONTEXT_FILE_LIMIT

    @context_file_limit.setter
    def context_file_limit(self, value):
        self._context_file_limit = value

    @property
    def forcerm(self):
        return self._forcerm if self._forcerm is not None else DEFAULT_FORCE_RM

    @forcerm.setter
    def forcerm(self, value):
        self._forcerm = value

    @property
    def infinite_command(self):
        return self._infinite_command or DEFAULT_INFINITE_COMMAND

    @infinite_command.setter
    def infinite_command(self, value):
        self._infinite_command = value

    @property
    def labels(self):
        return self._labels or {}

    @labels.setter
    def labels(self, value):
        self._labels = value

    @property
    def layers(self):
        return self._layers or []

    @layers.setter
    def layers(self, value):
        self._layers = value

    @property
    def mkdir_command(self):
        return self._mkdir_command or DEFAULT_MKDIR_COMMAND

    @mkdir_command.setter
    def mkdir_command(self, value):
        self._mkdir_command = value

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def pull(self):
        return self._pull if self._pull is not None else DEFAULT_PULL

    @pull.setter
    def pull(self, value):
        self._pull = value

    @property
    def rm(self):
        return self._rm if self._rm is not None else DEFAULT_RM

    @rm.setter
    def rm(self, value):
        self._rm = value

    @property
    def tags(self):
        return self._tags or []

    @tags.setter
    def tags(self, value):
        self.tags = value

    @property
    def volumes(self):
        """

        :return: Tuples with source:destination volume only for build.
        """
        return self._volumes or []

    @volumes.setter
    def volumes(self, value):
        self._volumes = value

    def update(self, buildargs=None,
               container_limits=None,
               context_file_limit=None,
               forcerm=None,
               infinite_command=None,
               labels=None,
               layers=None,
               mkdir_command=None,
               pull=None,
               rm=None,
               tags=None,
               volumes=None):
        self._buildargs = self._update_dict(self._buildargs, buildargs)
        self._container_limits = self._update_dict(self._container_limits, container_limits)
        self._context_file_limit = self._update_value(self._context_file_limit, context_file_limit)
        self._forcerm = self._update_value(self._forcerm, forcerm)
        self._infinite_command = self._update_value(self._infinite_command, infinite_command)
        self._labels = self._update_dict(self._labels, labels)
        self._layers = self._update_value(self._layers, layers)
        self._mkdir_command = self._update_value(self._mkdir_command, mkdir_command)
        self._pull = self._update_value(self._pull, pull)
        self._rm = self._update_value(self._rm, rm)
        self._tags = self._update_list(self._tags, tags)
        self._volumes = self._update_list(self._volumes, volumes)

    @staticmethod
    def _update_dict(origin, other):
        if (origin is None) and (other is None):
            return None

        result_dict = {}
        result_dict.update(origin or {})
        result_dict.update(other or {})
        return result_dict

    @staticmethod
    def _update_value(origin, other):
        return other if other is not None else origin

    @staticmethod
    def _update_list(origin, other):
        if (origin is None) and (other is None):
            return None
        result_list = origin or []

        if isinstance(other, six.string_types):
            result_list.append(other)
        elif other is not None:
            result_list += other

        return result_list

    @property
    def tags_and_repos(self):
        """
        :return: list with tupple (repo,tag)
        """
        result = []
        for t in self.tags:
            t_split = t.split(':')
            t_part_number = len(t_split)
            if t_part_number > 2:
                raise Exception("Wrong tag format.")
            elif t_part_number == 2:
                result.append((t_split[0], t_split[1]))
            elif t_part_number == 1:
                result.append((t, DEFAULT_TAG))
        return result

    @property
    def unused_buildargs(self):
        if self._unused_buildargs is None:
            self._unused_buildargs = set(self.buildargs.keys())
        return self._unused_buildargs

    def mark_used_buildarg(self, key):
        if key in self.unused_buildargs:
            self.unused_buildargs.remove(key)

    @staticmethod
    def validate_json_schema(config):
        """
        :return: True, if the config dictionary is valid. Else false.
        """

        try:
            name = config.get(CONFIG_NAME, "")
        except:
            name = ""

        try:
            version = config.get(CONFIG_VERSION, DEFAULT_VERSION)
            version_schema = schema.get(version)
            if not version_schema:
                raise Exception("Config version is not valid.")
            Draft3Validator(version_schema).validate(config)
            return True
        except jsonschema.ValidationError as ex:
            logger.warning("Config '{}' is not valid:\n{}".format(name, ex.message))
            return False
        except:
            logger.warning("Version for config '{}' is not valid.".format(name))
            return False

    @staticmethod
    def load_config_from_default_files():
        configs = []
        for f in DEFAULT_CONFIG_FILES:
            try:
                abs_path = os.path.normpath(os.path.abspath(f))
                if os.path.exists(abs_path):
                    logger.debug("Loading default config `{}`".format(f))
                    with open(f, mode='r') as config_file:
                        config_dict = json.load(config_file)
                        config_dict.setdefault(CONFIG_NAME, f)
                        configs.append(config_dict)
            except Exception as ex:
                message = "Cannot load default config file '{}':\n{}".format(f, ex.msg)
                logger.warning(message)

        return configs

    @staticmethod
    def merge_configs(config):
        config_object = ImageConfig()
        if config:
            if not isinstance(config, list):
                config = [config]
            for c in config:
                config_object += ImageConfig(config_dict=c)
        return config_object

    def __add__(self, other):
        first_dictionary = self.config
        second_dictionary = other.config

        new = ImageConfig(config_dict=first_dictionary)
        new.update(**second_dictionary)
        new.name = self.name + other.name
        return new
